{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# 测试"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch\n",
    "from torch import nn, Tensor\n",
    "from torch.nn import functional as F\n",
    "\n",
    "class FastGlobalAvgPool(nn.Module):\n",
    "    def __init__(self, flatten=False):\n",
    "        super().__init__()\n",
    "        self.flatten = flatten\n",
    "\n",
    "    def forward(self, x):\n",
    "        if self.flatten:\n",
    "            in_size = x.size()\n",
    "            return x.view((in_size[0], in_size[1], -1)).mean(dim=2)\n",
    "        else:\n",
    "            return x.view(x.size(0), x.size(1), -1).mean(-1).view(x.size(0), x.size(1), 1, 1)\n",
    "\n",
    "class BatchNorm(nn.BatchNorm2d):\n",
    "    def __init__(self, num_features, eps=1e-05, momentum=0.1, weight_freeze=False, bias_freeze=False, weight_init=1.0,\n",
    "                 bias_init=0.0):\n",
    "        super().__init__(num_features, eps=eps, momentum=momentum)\n",
    "        if weight_init is not None: nn.init.constant_(self.weight, weight_init)\n",
    "        if bias_init is not None: nn.init.constant_(self.bias, bias_init)\n",
    "        self.weight.requires_grad_(not weight_freeze)\n",
    "        self.bias.requires_grad_(not bias_freeze)\n",
    "\n",
    "class Head(nn.Module):\n",
    "    def __init__(self, feat_dim=1024, num_class=9):\n",
    "        super().__init__()\n",
    "        self.pool_layer = FastGlobalAvgPool()\n",
    "        self.bottleneck = BatchNorm(feat_dim, bias_freeze=True)\n",
    "        self.bnneck = nn.BatchNorm1d(num_class)\n",
    "        self.weight = nn.Parameter(Tensor(num_class, feat_dim))\n",
    "\n",
    "    def forward(self, x: Tensor) -> Tensor:\n",
    "        pool_feat = self.pool_layer(x)\n",
    "        neck_feat = self.bottleneck(pool_feat)\n",
    "        neck_feat = neck_feat.view(neck_feat.size(0), -1)\n",
    "        logits = F.linear(neck_feat,self.weight)\n",
    "        logits = self.bnneck(logits)\n",
    "        return logits"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "torch.cuda.empty_cache()\n",
    "model_path = \"/media/pc/data/board/arria10/lxw/tasks/tools/npu_user_demos/models/telecom_pt/Nin1_helmet_small/helmet_small.pth\"\n",
    "model = Head()\n",
    "state_dict = torch.load(model_path, weights_only=False, map_location=torch.device('cpu'))\n",
    "state_dict = {k.replace('bottleneck.0', 'bottleneck').replace('heads.', ''): v for k, v in state_dict['model'].items() if \"heads\" in k or \"bottleneck\" in k}\n",
    "model.load_state_dict(state_dict, strict=True)\n",
    "model = model.eval()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "import set_env\n",
    "import numpy as np\n",
    "import tvm\n",
    "from tvm import relay\n",
    "\n",
    "name = \"x\"\n",
    "shape = (1, 1024, 8, 6)\n",
    "data_np = (np.random.randint(0, 256, shape)/255).astype(\"float32\")\n",
    "data_torch = torch.from_numpy(data_np)\n",
    "scripted_model = torch.jit.trace(model, data_torch).eval()\n",
    "shape_list = [(name, shape)]\n",
    "mod, params = relay.frontend.from_pytorch(scripted_model, shape_list)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "fn (%x: Tensor[(1, 1024, 8, 6), float32] /* span=aten::size_0.x:0:0 */, %aten::batch_norm_0.weight: Tensor[(1024), float32] /* span=aten::batch_norm_0.weight:0:0 */, %aten::batch_norm_0.bias: Tensor[(1024), float32] /* span=aten::batch_norm_0.bias:0:0 */, %aten::batch_norm_0.running_mean: Tensor[(1024), float32] /* span=aten::batch_norm_0.running_mean:0:0 */, %aten::batch_norm_0.running_var: Tensor[(1024), float32] /* span=aten::batch_norm_0.running_var:0:0 */, %aten::linear_0.weight: Tensor[(9, 1024), float32] /* span=aten::linear_0.weight:0:0 */, %aten::batch_norm_1.weight: Tensor[(9), float32] /* span=aten::batch_norm_1.weight:0:0 */, %aten::batch_norm_1.bias: Tensor[(9), float32] /* span=aten::batch_norm_1.bias:0:0 */, %aten::batch_norm_1.running_mean: Tensor[(9), float32] /* span=aten::batch_norm_1.running_mean:0:0 */, %aten::batch_norm_1.running_var: Tensor[(9), float32] /* span=aten::batch_norm_1.running_var:0:0 */) {\n",
      "  %0 = reshape(%x, newshape=[1, 1024, -1]) /* span=aten::view_0:0:0 */;\n",
      "  %1 = mean(%0, axis=[-1]) /* span=aten::mean_0:0:0 */;\n",
      "  %2 = reshape(%1, newshape=[1, 1024, 1, 1]) /* span=aten::view_1:0:0 */;\n",
      "  %3 = nn.batch_norm(%2, %aten::batch_norm_0.weight, %aten::batch_norm_0.bias, %aten::batch_norm_0.running_mean, %aten::batch_norm_0.running_var) /* span=aten::batch_norm_0:0:0 */;\n",
      "  %4 = %3.0 /* span=aten::batch_norm_0:0:0 */;\n",
      "  %5 = reshape(%4, newshape=[1, -1]) /* span=aten::view_2:0:0 */;\n",
      "  %6 = nn.dense(%5, %aten::linear_0.weight, units=None) /* span=aten::linear_0:0:0 */;\n",
      "  %7 = nn.batch_norm(%6, %aten::batch_norm_1.weight, %aten::batch_norm_1.bias, %aten::batch_norm_1.running_mean, %aten::batch_norm_1.running_var) /* span=aten::batch_norm_1:0:0 */;\n",
      "  %7.0 /* span=aten::batch_norm_1:0:0 */\n",
      "}\n"
     ]
    }
   ],
   "source": [
    "print(mod[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "from tvm.relay.dataflow_pattern import rewrite\n",
    "from tvm_book.transforms.simplify import FastGlobalAvgPoolSimplify\n",
    "run_mod = tvm.IRModule()\n",
    "run_mod[\"main\"] = rewrite(FastGlobalAvgPoolSimplify(), mod[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<div class=\"highlight\" style=\"background: \"><pre style=\"line-height: 125%;\"><span></span><span style=\"color: #008000; font-weight: bold\">def</span> <span style=\"color: #AA22FF\">@main</span>(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x: Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">8</span>, <span style=\"color: #008000\">6</span>), float32] span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::size_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>x:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight: Tensor[(<span style=\"color: #008000\">1024</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias: Tensor[(<span style=\"color: #008000\">1024</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean: Tensor[(<span style=\"color: #008000\">1024</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var: Tensor[(<span style=\"color: #008000\">1024</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::linear_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight: Tensor[(<span style=\"color: #008000\">9</span>, <span style=\"color: #008000\">1024</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::linear_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight: Tensor[(<span style=\"color: #008000\">9</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias: Tensor[(<span style=\"color: #008000\">9</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean: Tensor[(<span style=\"color: #008000\">9</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var: Tensor[(<span style=\"color: #008000\">9</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>) {\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>adaptive_avg_pool2d(<span style=\"color: #AA22FF; font-weight: bold\">%</span>x, output_size<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> ty<span style=\"color: #AA22FF; font-weight: bold\">=</span>Tensor[(<span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1024</span>, <span style=\"color: #008000\">1</span>, <span style=\"color: #008000\">1</span>), float32] <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>batch_norm(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">0</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">1.0</span> <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_0:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> reshape(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">2</span>, newshape<span style=\"color: #AA22FF; font-weight: bold\">=</span>[<span style=\"color: #008000\">1</span>, <span style=\"color: #AA22FF; font-weight: bold\">-</span><span style=\"color: #008000\">1</span>]) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::view_2:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>dense(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">3</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::linear_0<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight, units<span style=\"color: #AA22FF; font-weight: bold\">=</span><span style=\"color: #008000; font-weight: bold\">None</span>) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::linear_0:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5</span> <span style=\"color: #AA22FF; font-weight: bold\">=</span> nn<span style=\"color: #AA22FF; font-weight: bold\">.</span>batch_norm(<span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">4</span>, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>weight, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>bias, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_mean, <span style=\"color: #AA22FF; font-weight: bold\">%</span>aten::batch_norm_1<span style=\"color: #AA22FF; font-weight: bold\">.</span>running_var) <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>;\n",
       "  <span style=\"color: #AA22FF; font-weight: bold\">%</span><span style=\"color: #008000\">5.0</span> <span style=\"color: #AA22FF; font-weight: bold\">/*</span> span<span style=\"color: #AA22FF; font-weight: bold\">=</span>aten::batch_norm_1:<span style=\"color: #008000\">0</span>:<span style=\"color: #008000\">0</span> <span style=\"color: #AA22FF; font-weight: bold\">*/</span>\n",
       "}\n",
       "</pre></div>\n"
      ],
      "text/plain": [
       "<IPython.core.display.HTML object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "run_mod.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [
    {
     "name": "stderr",
     "output_type": "stream",
     "text": [
      "One or more operators have not been tuned. Please tune your model for better performance. Use DEBUG logging level to see more details.\n"
     ]
    }
   ],
   "source": [
    "with tvm.transform.PassContext(opt_level=3, disabled_pass={\"AlterOpLayout\"}):\n",
    "    lib = relay.build(mod, target=\"llvm\", params=params)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(tvm.cpu(0)))\n",
    "module.run(**{name: data_np})\n",
    "num_outputs = module.get_num_outputs()\n",
    "origin_outputs = [module.get_output(k).numpy() for k in range(num_outputs)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "with tvm.transform.PassContext(opt_level=3, disabled_pass={\"AlterOpLayout\"}):\n",
    "    lib = relay.build(run_mod, target=\"llvm\", params=params)\n",
    "func = lib[lib.libmod_name]\n",
    "module = tvm.contrib.graph_executor.GraphModule(func(tvm.cpu(0)))\n",
    "module.run(**{name: data_np})\n",
    "num_outputs = module.get_num_outputs()\n",
    "outputs = [module.get_output(k).numpy() for k in range(num_outputs)]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "np.testing.assert_allclose(origin_outputs[0], outputs[0])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 26,
   "metadata": {},
   "outputs": [],
   "source": [
    "from copy import deepcopy\n",
    "import tvm\n",
    "from tvm import relay\n",
    "from tvm.relay.quantize.quantize import _bind_params\n",
    "optimize = tvm.transform.Sequential(\n",
    "    [\n",
    "        relay.transform.SimplifyInference(),\n",
    "        relay.transform.FoldConstant(),\n",
    "        relay.transform.FoldScaleAxis(),\n",
    "        # relay.transform.CanonicalizeOps(),\n",
    "        # relay.transform.FoldConstant(),\n",
    "    ]\n",
    ")\n",
    "run_mod = deepcopy(mod)\n",
    "run_mod[\"main\"] = rewrite(FastGlobalAvgPoolSimplify(), run_mod[\"main\"])\n",
    "run_mod[\"main\"] = _bind_params(run_mod[\"main\"], params)\n",
    "with tvm.transform.PassContext(opt_level=3):\n",
    "    run_mod2 = relay.quantize.prerequisite_optimize(deepcopy(run_mod), params)\n",
    "    run_mod = optimize(run_mod)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 27,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "fn (%x: Tensor[(1, 1024, 8, 6), float32] /* ty=Tensor[(1, 1024, 8, 6), float32] span=aten::size_0.x:0:0 */) -> Tensor[(1, 9), float32] {\n",
      "  %0 = nn.adaptive_avg_pool2d(%x, output_size=[1, 1]) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %1 = multiply(%0, meta[relay.Constant][0] /* ty=Tensor[(1024, 1, 1), float32] */) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %2 = add(%1, meta[relay.Constant][1] /* ty=Tensor[(1024, 1, 1), float32] */) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %3 = reshape(%2, newshape=[1, -1]) /* ty=Tensor[(1, 1024), float32] span=aten::view_2:0:0 */;\n",
      "  %4 = nn.dense(%3, meta[relay.Constant][2] /* ty=Tensor[(9, 1024), float32] */, units=None) /* ty=Tensor[(1, 9), float32] */;\n",
      "  add(%4, meta[relay.Constant][3] /* ty=Tensor[(9), float32] */) /* ty=Tensor[(1, 9), float32] */\n",
      "} /* ty=fn (Tensor[(1, 1024, 8, 6), float32]) -> Tensor[(1, 9), float32] */\n",
      "\n"
     ]
    }
   ],
   "source": [
    "print(run_mod[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "fn (%x: Tensor[(1, 1024, 8, 6), float32] /* ty=Tensor[(1, 1024, 8, 6), float32] span=aten::size_0.x:0:0 */) -> Tensor[(1, 9), float32] {\n",
      "  %0 = nn.adaptive_avg_pool2d(%x, output_size=[1, 1]) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %1 = multiply(%0, meta[relay.Constant][0] /* ty=Tensor[(1024, 1, 1), float32] */) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %2 = add(%1, meta[relay.Constant][1] /* ty=Tensor[(1024, 1, 1), float32] */) /* ty=Tensor[(1, 1024, 1, 1), float32] */;\n",
      "  %3 = reshape(%2, newshape=[1, -1]) /* ty=Tensor[(1, 1024), float32] span=aten::view_2:0:0 */;\n",
      "  %4 = nn.dense(%3, meta[relay.Constant][2] /* ty=Tensor[(9, 1024), float32] */, units=None) /* ty=Tensor[(1, 9), float32] */;\n",
      "  add(%4, meta[relay.Constant][3] /* ty=Tensor[(9), float32] */) /* ty=Tensor[(1, 9), float32] */\n",
      "} /* ty=fn (Tensor[(1, 1024, 8, 6), float32]) -> Tensor[(1, 9), float32] */\n",
      "\n"
     ]
    }
   ],
   "source": [
    "print(run_mod2[\"main\"])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "ai",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.4"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
